{-# LANGUAGE DataKinds           #-}
{-# LANGUAGE GADTs               #-}
{-# LANGUAGE KindSignatures      #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE StandaloneDeriving  #-}
module Data.FixedWidth.Word (FixedWord) where

import           Data.Bifunctor
import           Data.Bits
import           Data.FixedWidth.Shared
import           Data.Kind
import           Data.Proxy
import           Data.Ratio
import           Data.Word
import           GHC.TypeLits

data FixedWord :: Nat -> Type where
    FixedWord :: KnownNat n => Word64 -> FixedWord n

deriving instance Eq (FixedWord n)
deriving instance Ord (FixedWord n)

instance Show (FixedWord n) where
    show (FixedWord n) = show n

instance (KnownNat n) => Bounded (FixedWord n) where
    minBound = FixedWord 0
    maxBound = FixedWord $ 2 ^ width - 1
        where width = bitWidth (FixedWord 0 :: FixedWord n)

instance (KnownNat n) => Enum (FixedWord n) where
    toEnum v
        | FixedWord v >= (minBound :: FixedWord n) && FixedWord v <= (maxBound :: FixedWord n) = FixedWord v
        | otherwise = error $ concat ["\"", show v, "\" is too large to fit in an ", show $ bitWidth (FixedWord 0 :: FixedWord n), "-bit integer"]
    fromEnum (FixedWord v) = v

instance (KnownNat n) => Num (FixedWord n) where
    (FixedWord a) + (FixedWord b) = FixedWord $ if res > max then res .&. max - (max + 1) else res
        where
            res = a + b
            (FixedWord min) = minBound :: FixedWord n
            (FixedWord max) = maxBound :: FixedWord n
    (FixedWord a) * (FixedWord b) = FixedWord $ if res > max || res < min then res .&. max - (max + 1) else res
        where
            res = a * b
            (FixedWord min) = minBound :: FixedWord n
            (FixedWord max) = maxBound :: FixedWord n
    (FixedWord a) - (FixedWord b) = FixedWord $ if res < min then res .&. max - (max + 1) else res
        where
            res = a - b
            (FixedWord min) = minBound :: FixedWord n
            (FixedWord max) = maxBound :: FixedWord n
    negate f@(FixedWord a) = if f == minBound then minBound else FixedWord $ negate a
    abs f@(FixedWord a) = if f == minBound then minBound else FixedWord $ abs a
    signum (FixedWord a) = FixedWord $ signum a
    fromInteger int
        | int >= min' && int <= max' = FixedWord $ fromInteger int
        | otherwise = error $ concat ["\"", show int, "\" is too large to fit in an ", show $ bitWidth (FixedWord 0 :: FixedWord n), "-bit integer"]
        where
            (FixedWord min) = minBound :: FixedWord n
            (FixedWord max) = maxBound :: FixedWord n
            min' = toInteger min
            max' = toInteger max

instance (KnownNat n) => Real (FixedWord n) where
    toRational (FixedWord a) = (toInteger a) % 1

instance (KnownNat n) => Integral (FixedWord n) where
    (FixedWord a) `quotRem` (FixedWord b) = bimap FixedWord FixedWord $ quotRem a b
    toInteger (FixedWord a) = toInteger a
